from textwrap import dedent
from typing import List, Optional

from opennem.api.stats.schema import ScadaDateRange
from opennem.core.normalizers import normalize_duid
from opennem.schema.network import NetworkNEM, NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod


def station_id_case(station_codes: List[str]) -> str:
    return ",".join(
        ["'{}'".format(i) for i in map(normalize_duid, station_codes)]
    )


def observation_query(
    station_codes: List[str],
    interval: TimeInterval,
    network: NetworkSchema = NetworkNEM,
    timezone: str = "UTC",
    period: Optional[TimePeriod] = None,
    scada_range: Optional[ScadaDateRange] = None,
    year: Optional[str] = None,
):

    timezone = network.get_timezone(postgres_format=True)

    if not timezone:
        timezone = "UTC"

    __query = """
    SET SESSION TIME ZONE '{timezone}';

    select
        time_bucket_gapfill('{trunc}', observation_time) as observation_time,
        fs.station_id,
        interpolate(avg(fs.temp_air)) as temp_avg,
        interpolate(min(fs.temp_air)) as temp_min,
        interpolate(max(fs.temp_air)) as temp_max
    from bom_observation fs
    where
        fs.station_id in ({station_codes})
        and fs.observation_time <= {date_end}
        and fs.observation_time >= {date_start_condition}
    group by 1, 2
    order by 1 desc, 2 desc
    """

    date_end = "now()"

    if scada_range:
        date_end = scada_range.get_end_sql()

    date_start_condition = ""

    if period:
        date_start_condition = "{date_end}::timestamp - '{period}'::interval".format(
            date_end=date_end, period=period.period_sql
        )

    if year:
        date_start_condition = "'{year}-01-01'::date".format(year=year)

    if not period and not year:
        date_start_condition = "'{}'".format(scada_range.get_start())

    query = dedent(
        __query.format(
            station_codes=station_id_case(station_codes),
            trunc=interval.interval_sql,
            timezone=timezone,
            date_start_condition=date_start_condition,
            date_end=date_end,
        )
    )

    return query


def observation_query_all(
    station_codes: List[str],
    scada_range: ScadaDateRange,
    network: NetworkSchema = NetworkNEM,
):

    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """
        select
            date_trunc('month', t.observation_time at time zone '{timezone}') as observation_month,
            t.station_id,
            avg(t.temp_avg),
            min(t.temp_min),
            max(t.temp_max)
        from
            (
                select
                    time_bucket_gapfill('1 day', observation_time) as observation_time,
                    fs.station_id,
                    interpolate(avg(fs.temp_air)) as temp_avg,
                    interpolate(min(fs.temp_air)) as temp_min,
                    interpolate(max(fs.temp_air)) as temp_max
                from bom_observation fs
                where
                    fs.station_id in ({station_codes})
                    and fs.observation_time <= {date_end}
                    and fs.observation_time >= {date_start}
                group by 1, 2
            ) as t
        group by 1, 2
        order by 1 desc, 2 desc
    """

    date_end = scada_range.get_end_sql()
    date_start = scada_range.get_start_sql()

    query = __query.format(
        station_codes=station_id_case(station_codes),
        timezone=timezone,
        date_start=date_start,
        date_end=date_end,
    )

    return query




import os
from typing import Optional

import yaml

LOGGING_CONFIG = {}


SETTINGS_DIR = os.path.realpath(os.path.dirname(__file__))


class SettingsNotFound(Exception):
    pass


def load_logging_config(
    filename: str = "logging.yml", fail_silent: bool = True
) -> Optional[dict]:
    """


    """

    settings_filepath = os.path.realpath(os.path.join(SETTINGS_DIR, filename))

    if not os.path.isfile(settings_filepath):
        if fail_silent:
            return dict()

        raise SettingsNotFound(
            "Not a valid logging settings file: {}".format(settings_filepath)
        )

    with open(settings_filepath, "rt") as f:
        config_data = yaml.safe_load(f.read())

    return config_data


def get_logging_config():
    return LOGGING_CONFIG


LOGGING_CONFIG = load_logging_config()



"""Update on_energy_sum to support variable interval sizes, custom tenergy type, and calculate using built-in polygon functions

Revision ID: 4fb955a7a2a7
Revises: bb6fb4645ca3
Create Date: 2020-10-21 17:35:48.326616

"""
from alembic import op

from opennem.core.loader import load_data

# revision identifiers, used by Alembic.
revision = "4fb955a7a2a7"
down_revision = "bb6fb4645ca3"
branch_labels = None
depends_on = None


def upgrade():
    ENERGY_FUNCTIONS = load_data("energy_methods.sql", from_fixture=True)

    op.execute(
        "drop aggregate if exists on_energy_sum (numeric, text) cascade"
    )
    op.execute(
        "drop function if exists agg_power(agg_state numeric[], el numeric) cascade"
    )
    op.execute(
        "drop function if exists onenergy_finalfunc(agg_state numeric[])"
    )

    op.execute(ENERGY_FUNCTIONS)


def downgrade():
    op.execute(
        "drop aggregate if exists on_energy_sum (numeric, text) cascade"
    )
    op.execute("drop type if exists tenergy cascade")



import csv
import json
import logging
from datetime import datetime
from functools import reduce
from itertools import groupby
from pathlib import Path
from pprint import pprint
from typing import Optional

from opennem.core.facilitystatus import parse_facility_status
from opennem.core.loader import load_data
from opennem.core.normalizers import station_name_cleaner
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.opennem import StationSchema
from opennem.schema.stations import StationSet

logger = logging.getLogger("opennem.importer.mms")


def parse_mms_date(date_string: str) -> Optional[datetime]:
    """

        `25/10/1998  12:00:00 am` => d

    """
    date_string_components = date_string.strip().split(" ")

    if len(date_string_components) < 2:
        raise Exception("Error parsing date: {}".format(date_string))

    date_part = date_string_components[0]

    dt = None

    try:
        dt = datetime.strptime(date_part, "%Y/%m/%d")
    except ValueError:
        raise Exception("Error parsing date: {}".format(date_string))

    # AEMO sets dates to this value when they mean None
    # if dt.year == 2999:
    # return None

    return dt


def dudetailsummary_grouper(tables):

    if "PARTICIPANT_REGISTRATION_DUDETAILSUMMARY" not in tables:
        raise Exception("No dudetailsummary table")

    records = tables["PARTICIPANT_REGISTRATION_DUDETAILSUMMARY"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "date_start": parse_mms_date(i["START_DATE"]),
            "date_end": parse_mms_date(i["END_DATE"]),
            # "status": "operating",
            "code": i["DUID"],
            "network_region": i["REGIONID"],
            "station_code": i["STATIONID"],
            "participant_id": i["PARTICIPANTID"],
            "dispatch_type": i["DISPATCHTYPE"],
        }
        for i in records
    ]

    now = datetime.now()

    records = [
        {
            **i,
            "status": parse_facility_status("retired")
            if i["date_end"] <= now
            else parse_facility_status("operating"),
        }
        for i in records
    ]

    grouped_records = {}

    # First pass sorts facilities into stations
    for k, v in groupby(records, lambda x: (x["station_code"], x["code"])):
        key = k[0]
        duid = k[1]
        if key not in grouped_records:
            grouped_records[key] = {}
            grouped_records[key]["station_code"] = k[0]
            # grouped_records[key]["participant"] = v[0]["PARTICIPANTID"]
            grouped_records[key]["details"] = {}
            grouped_records[key]["facilities"] = []

        if duid not in grouped_records[key]["details"]:
            grouped_records[key]["details"][duid] = []

        grouped_records[key]["details"][duid] += list(v)

    # Second pass flatten the records and we should get start and end dates
    # and a derived status
    for rec in grouped_records.keys():
        for _, facility_group_records in grouped_records[rec][
            "details"
        ].items():

            # date_end_min = min(
            #     facility_group_records, key=lambda x: x["date_end"]
            # )
            date_end_max = max(
                facility_group_records, key=lambda x: x["date_end"]
            )
            date_start_min = min(
                facility_group_records, key=lambda x: x["date_start"]
            )

            grouped_rec = {
                **date_end_max,
                "date_start": date_start_min["date_start"],
            }

            if grouped_rec["date_end"].year == 2999:
                grouped_rec["date_end"] = None

            grouped_records[rec]["facilities"].append(grouped_rec)

    grouped_records = [
        {"station_code": i, "facilities": v["facilities"]}
        for i, v in grouped_records.items()
    ]

    tables["PARTICIPANT_REGISTRATION_DUDETAILSUMMARY"] = grouped_records

    for record in grouped_records:
        station_code = record["station_code"]

        if station_code not in mms:
            print("dudetailsummary: {} not in mms".format(station_code))
            continue

        # mms[station_code] = {**record, **mms[station_code]}
        mms[station_code]["facilities"] = record["facilities"]

    tables["mms"] = mms

    return tables


def operatingstatus_grouper(tables):

    if "PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS" not in tables:
        raise Exception(
            "No PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS table"
        )

    records = tables["PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "effective_date": parse_mms_date(i["EFFECTIVEDATE"]),
            "station_code": i["STATIONID"],
            "status": i["STATUS"],
        }
        for i in records
    ]

    grouped_records = {}

    for station_code, records in groupby(records, lambda x: x["station_code"]):
        if station_code not in grouped_records:
            grouped_records[station_code] = {}
            grouped_records[station_code]["id"] = station_code
            grouped_records[station_code]["details"] = []

        grouped_records[station_code]["details"] += list(records)

    for station_code, record in grouped_records.items():
        date_max = max(record["details"], key=lambda x: x["effective_date"])
        grouped_records[station_code]["status"] = date_max["status"]
        # grouped_records[station_code]["status"] = date_max["status"]

        if station_code not in mms:
            print("operatingstatus: {} is not in MMS".format(station_code))

        mms[station_code]["status"] = date_max["status"]

    tables["PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS"] = grouped_records

    tables["mms"] = mms

    return tables


def stations_grouper(tables):

    if "PARTICIPANT_REGISTRATION_STATION" not in tables:
        raise Exception("No PARTICIPANT_REGISTRATION_STATION table")

    records = tables["PARTICIPANT_REGISTRATION_STATION"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "id": _id,
            "updated_at": parse_mms_date(i["LASTCHANGED"]),
            "name": station_name_cleaner(i["STATIONNAME"]),
            "code": i["STATIONID"],
            "station_code": i["STATIONID"],
            "network_name": i["STATIONNAME"],
            "address1": i["ADDRESS1"],
            "address2": i["ADDRESS2"],
            "locality": i["CITY"],
            "state": i["STATE"],
            "postcode": i["POSTCODE"],
            "facilities": [],
        }
        for _id, i in enumerate(records, start=1000)
    ]

    for record in records:
        station_code = record["station_code"]

        if station_code not in mms:
            mms[station_code] = {}

        mms[station_code] = record

    tables["mms"] = mms

    return tables


def dudetail_grouper(tables):

    if "PARTICIPANT_REGISTRATION_DUDETAIL" not in tables:
        raise Exception("No PARTICIPANT_REGISTRATION_DUDETAIL table")

    records = tables["PARTICIPANT_REGISTRATION_DUDETAIL"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "code": i["DUID"],
            "version": i["VERSIONNO"],
            "capacity_registered": float(i["REGISTEREDCAPACITY"]),
            "capacity_maximum": float(i["MAXCAPACITY"]),
            "dispatch_type": i["DISPATCHTYPE"],
        }
        for _id, i in enumerate(records, start=1000)
    ]

    grouped_records = {}

    for network_code, records in groupby(records, lambda x: x["code"]):
        if network_code not in grouped_records:
            grouped_records[network_code] = {}
            grouped_records[network_code]["details"] = []

        grouped_records[network_code]["details"] += list(records)

    for network_code, record in grouped_records.items():
        version_max = max(record["details"], key=lambda x: x["version"])

        dudetail_record = {
            "capacity_registered": version_max["capacity_registered"],
            "capacity_maximum": version_max["capacity_maximum"],
        }

        facility = None

        for station_code in mms.keys():

            for fac in mms[station_code]["facilities"]:
                if fac["code"] == network_code:

                    # don't need the version field any more
                    fac.pop("version", None)

                    facility = {
                        **fac,
                        **dudetail_record,
                    }
                    fac_index = mms[station_code]["facilities"].index(fac)
                    mms[station_code]["facilities"][fac_index] = facility

        if not facility:
            print("dudetail: couldn't find facility: {}".format(network_code))

    tables["PARTICIPANT_REGISTRATION_DUDETAIL"] = grouped_records

    tables["mms"] = mms

    return tables


def load_aemo_csv(item, filename):

    if not item:
        item = {}

    if type(item) is not dict:
        raise Exception(
            "Invalid item type expecting a dict so we can fill it "
        )

    current_item = load_data(filename, True)

    if "content" not in current_item:
        logger.error("No content in item to parse")
        return item

    content = current_item["content"]
    del current_item["content"]

    table_name = None
    table_values = ""
    table_fields = []
    table_records = []

    content_split = content.splitlines()

    datacsv = csv.reader(content_split)

    for row in datacsv:
        if not row or type(row) is not list or len(row) < 1:
            continue

        record_type = row[0]

        if record_type == "C":
            # @TODO csv meta stored in table
            if table_name is not None:
                item[table_name] = table_records

        elif record_type == "I":
            if table_name is not None:
                item[table_name] = table_records

            table_name = "{}_{}".format(row[1], row[2])
            table_fields = row[4:]
            table_records = []

        elif record_type == "D":
            table_values = row[4:]
            record = dict(zip(table_fields, table_values))
            table_records.append(record)

    return item


def load_mms_tables():
    mms_path = Path(__file__).parent.parent / "data" / "mms"

    mms_files = [
        "mms/{}".format(f.name)
        for f in mms_path.iterdir()
        if f.suffix in [".zip"]
    ]

    tables = reduce(load_aemo_csv, mms_files, {})

    return tables


def mms_import():
    tables = load_mms_tables()

    logger.info(
        "Imported {} tables: {}".format(
            len(tables.keys()), ", ".join(list(tables.keys()))
        )
    )

    tables = stations_grouper(tables)
    tables = dudetailsummary_grouper(tables)
    tables = operatingstatus_grouper(tables)
    tables = dudetail_grouper(tables)

    _mms = tables["mms"]

    mms = StationSet()

    for s in _mms.values():
        mms.add_dict(s)

    return mms


def mms_station_map_from_records(mms):
    """
        Get the station to duid map from MMS and return it
    """

    mms_duid_station_map = {}

    for station_record in mms.as_list():
        for network_code in [i.code for i in station_record.facilities]:
            mms_duid_station_map[network_code] = station_record.code

    return mms_duid_station_map


def mms_export():
    """

        Export MMS records

        @TODO move this to opennem.export and keep modules consistent
    """
    mms = mms_import()

    mms_duid_station_map = mms_station_map_from_records(mms)

    with open("data/mms.json", "w") as fh:
        fh.write(mms.json(indent=4))

    with open("data/mms_duid_station_map.json", "w") as fh:
        json.dump(mms_duid_station_map, fh, indent=4, cls=OpenNEMJSONEncoder)


if __name__ == "__main__":
    mms_export()



import re
from decimal import Decimal
from typing import Optional, Union

from opennem.core.station_names import station_map_name

__id_unit_regex = re.compile(r"^(Y|C)[0-9]{1,3}")

__id_duid_regex = re.compile(r"[A-Z0-9]{1,6}")


STRIP_WORDS = [
    "stage 1",
    "stage 2",
    "stage 3",
    "stage 4",
    "stage 5",
    "phase 1",
    "phase 2",
    "phase 3",
    "phase 4",
    "phase 5",
    "energy facility",
    "coal mine",
    "nett off",
    "renewable energy facility",
    "waste disposal facility",
    "network support station",
    "combined cycle",
    "green power hub units",
    "wind and solar",
    "streams station",
    "utilisation facility",
    "gas turbines",
    "stockland development",
    "backup generation",
    "gas turbine",
    "mini hydro",
    "power hub",
    "wastewater treatment",
    "water treatment",
    "wind farm",
    "solar park",
    "solar farm",
    "solar system",
    "power station",
    "bio reactor",
    "generation station",
    "sydney",
    "cogeneration",
    "kograh",
    "technologies",
    "biomass",
    "site",
    "project",
    "landfill",
    "wwt",
    "waste",
    "ltd",
    "agl",
    "pty",
    "units",
    "unit",
    "gas",
    "hydro",
    "diesel",
    "turbine",
    "system",
    "plant",
    "battery",
    "power",
    "farm",
    "sv",
    "sf",
    "pe",
    "pf",
    "pv",
    "ps",
    "solar",
    "storage",
    "wind",
    "ccgt",
    "renewable",
    "water",
    "lfg",
    "nsw",
    "vic",
    "qld",
    "sa",
    "dhl6",
    "run of river",
    "and",
    "bess",
    "h2e",
    "bess1",
    "gen",
    "gt",
    "bioreactor",
    "bordertown",
]

ACRONYMS = [
    "cec",
    "bhp",
    "rsl",
    "csu",
    "dhp",
    "stp",
    "wtp",
    "hvdc",
    "scs",
    "sips",
    "jv",
    "gt",
    "lmcc",
    "sa",
    "vpp",
    "fpc",
    "cd",
    "ab",
    "stp",
    "krc",
    "h2e",
    "jl",
    "uom",
    "agl",
    "nawma",
]


__is_number = re.compile(r"^[\d\.]+$")
__is_single_number = re.compile(r"^\d$")

# @TODO put all these helpers in utils/


def strip_whitespace(subject: str) -> str:
    return str(re.sub(r"\s+", "", subject.strip()))


def normalize_whitespace(subject: str) -> str:
    return str(re.sub(r"\s{2,}", " ", subject.strip()))


def is_number(value: Union[str, int]) -> bool:
    if type(value) is int:
        return True

    value = str(value)

    if re.match(__is_number, value):
        return True

    return False


def is_single_number(value: Union[str, int]) -> bool:
    if re.match(__is_single_number, value):
        return True
    return False


# @TODO not yet implemented
def id_unit(facility_name):
    __n = facility_name

    return __n


# @TODO not yet implement
def id_duid(facility_name):
    __n = facility_name

    return __n


def normalize_aemo_region(region_code: str = "") -> str:
    region_code = region_code.strip().upper()

    return region_code


def normalize_duid(duid: str) -> str:
    duid = duid or ""

    # @TODO replace with regexp that removes junk
    duid = duid.strip()

    if duid == "-":
        return ""

    return duid


def normalize_string(csv_string: str) -> str:
    if not type(csv_string) is str:
        csv_string = str(csv_string)

    csv_string = csv_string.strip().title()

    return csv_string


def name_normalizer(name: str) -> str:
    name_normalized = None

    if name and type(name) is str:
        name_normalized = name.strip()

    name_normalized = re.sub(r"\-", "", name_normalized)

    return str(name_normalized)


def clean_float(number: Union[str, int, float]) -> Optional[float]:
    num_return = None

    if isinstance(number, str):
        number = number.replace("-", "").strip()

        if number == "":
            return None

        num_return = float(number)
        return num_return

    if isinstance(number, int):
        return float(number)

    if isinstance(number, Decimal):
        return float(number)

    if isinstance(number, float):
        return number

    return None


def clean_numbers(part: Union[str, int]) -> Union[str, int, None]:
    """
        Clean the number part of a station name

    """
    if not is_number(part):
        return part

    part_parsed = int(part)

    if part_parsed < 6:
        return part_parsed

    return None


def station_name_cleaner(facility_name: str) -> str:
    """"
        This cleans station names from their messy as hell AEMO names to something
        we can plug into name_clean and humans can actually read

        It's a bit of a mess and could use a refactor


    """
    name_clean = facility_name or ""

    if type(facility_name) is str:
        name_clean = facility_name.strip()
    else:
        name_clean = str(facility_name).strip()

    # @todo check has duid / unit other

    name_clean = name_clean.lower()

    # @TODO replace with the re character stripped - this is unicode junk
    name_clean = name_clean.replace("\u00a0", " ")

    if station_map_name(name_clean) != name_clean:
        return station_map_name(name_clean)

    # strip units from name
    name_clean = re.sub(r"\d+\ ?(mw|kw|MW|KW)", "", name_clean)

    # strip other chars
    name_clean = re.sub(r",|-|\(|\)|\–|\"|\'", "", name_clean)
    # name_clean = re.sub(r"(\W|\ )+", "", name_clean)

    name_clean = re.sub(" +", " ", name_clean)

    name_clean = name_clean.replace("yalumba winery", "yalumba")
    name_clean = name_clean.replace("university of melbourne", "uom")

    # @TODO remove these hard codes
    if not name_clean in ["barcaldine solar farm", "Darling Downs Solar Farm"]:
        for w in STRIP_WORDS:
            if name_clean.startswith("todae solar") and w in ["solar"]:
                continue

            if " " in w:
                name_clean = name_clean.replace(w, "")

    name_components = [str(i) for i in name_clean.strip().split(" ")]
    name_components_parsed = []

    for comp in name_components:
        comp = str(comp)
        comp = comp.strip()
        comp = re.sub(r",|-|\(|\)|\–", "", comp)

        if type(comp) is not str:
            comp = None

        if comp == "":
            comp = None

        if comp in STRIP_WORDS:
            comp = None

        if comp in ACRONYMS:
            comp = comp.upper()
        elif type(comp) is str and comp.startswith("mc"):
            comp = "Mc" + comp[2:].capitalize()
        elif type(comp) is str and comp != "":
            comp = comp.capitalize()

        # strip numbers greater than 5
        comp = clean_numbers(comp)

        name_components_parsed.append(comp)

    if name_components_parsed[0] == "uom":
        name_components_parsed = name_components_parsed[:-1]

    name_clean = " ".join(
        [str(i) for i in name_components_parsed if i is not None]
    )

    name_clean = re.sub(" +", " ", name_clean)

    name_clean = name_clean.strip()

    if "/" in name_clean:
        name_clean = " / ".join(
            [i.strip().title() for i in name_clean.split("/")]
        )

    if station_map_name(name_clean) != name_clean:
        return station_map_name(name_clean)

    # uom special case
    name_clean = name_clean.replace("UOM ", "UoM ")

    # todae special case
    todae_match = re.match(r"^(Todae)\ (.*)", name_clean)
    if todae_match:
        todae_name, todae_rest = todae_match.groups()

        name_clean = "{} ({})".format(todae_rest, todae_name)

    return name_clean


def participant_name_filter(participant_name: str) -> Optional[str]:
    participant_name = strip_whitespace(participant_name)

    _p = (
        participant_name.strip()
        .replace("Pty Ltd", "")
        .replace("Ltd", "")
        .replace("/", " / ")
    )

    _p = re.sub(" +", " ", _p).strip()

    _p = _p.strip()

    if _p == "":
        return None

    return _p


def clean_capacity(
    capacity: Union[str, int, float], round_to: int = 6
) -> Optional[float]:
    """
        Takes a capacity and cleans it up into a float

        @TODO support unit conversion
    """
    cap_clean = None

    if capacity is None:
        return None

    # Type gating float, string, int, others ..
    if type(capacity) is float:
        cap_clean = capacity

    elif type(capacity) == str:
        cap_clean = strip_whitespace(capacity)

        if "-" in cap_clean:
            cap_clean_components = cap_clean.split("-")
            cap_clean = cap_clean_components.pop()
            return clean_capacity(cap_clean)

        if cap_clean == "":
            return None

        # funky values in spreadsheet
        cap_clean = cap_clean.replace(",", ".")
        cap_clean = float(cap_clean)

    elif type(capacity) == int:
        cap_clean = float(capacity)

    elif type(capacity) is not float:
        if capacity is None:
            return None

        raise Exception(
            "Capacity clean of type {} not supported: {}".format(
                type(capacity), capacity
            )
        )

    if cap_clean is None:
        return None

    cap_clean = round(cap_clean, round_to)

    return cap_clean



import os
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

BASE_ENV_FILE_NAME = ".env"


def replace_database_in_url(db_url, db_name):
    """
        replaces the database portion of a database connection URL with db_name

        @param db_name database name to replace with
    """

    db_url_parsed = urlparse(db_url)
    db_url_parsed = db_url_parsed._replace(path=f"/{db_name}")

    return db_url_parsed.geturl()


def load_env_file(environment: str) -> Optional[Path]:
    environment_suffix = environment.strip().lower()

    environment_env_file = "{}.{}".format(
        BASE_ENV_FILE_NAME, environment_suffix
    )

    env_path = Path(os.getcwd()) / Path(environment_env_file)

    if env_path.is_file():
        return env_path

    return None



"""observation_time index and not nullable

Revision ID: b16d7153a221
Revises: 637faac3f9c1
Create Date: 2020-10-22 00:54:37.709402

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "b16d7153a221"
down_revision = "637faac3f9c1"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        op.f("ix_bom_observation_observation_time"),
        "bom_observation",
        ["observation_time"],
        unique=False,
    )
    op.drop_index(
        "bom_observation_observation_time_idx", table_name="bom_observation"
    )


def downgrade():
    op.create_index(
        "bom_observation_observation_time_idx",
        "bom_observation",
        ["observation_time"],
        unique=False,
    )
    op.drop_index(
        op.f("ix_bom_observation_observation_time"),
        table_name="bom_observation",
    )



# pylint: skip-file
# python3
from io import BytesIO, StringIO
from zipfile import BadZipfile, ZipFile

import pandas as pd


class ZipFileStreamer(ZipFile):
    """ZipFile subclass, with method to extract ZipFile as byte or string stream to memory"""

    def __init__(
        self,
        filename="/data/marble/nemweb/ARCHIVE/DispatchIS_Reports/PUBLIC_DISPATCHIS_20170208.zip",
    ):
        """Open the ZIP file with mode read "r", write "w" or append "a" """
        ZipFile.__init__(self, filename)
        self.member_count = len(self.filelist)

    def extract_stream(self, member, IO=BytesIO):
        """	Extract a member from the archive as a byte stream or string
        steam, using its full name. `member' may be a filename or a
        ZipInfo object. """
        return IO(self.read(member))


def zip_streams(
    fileobject="/data/marble/nemweb/ARCHIVE/DispatchIS_Reports/PUBLIC_DISPATCHIS_20140208.zip",
):
    """Generator that yields each memeber of a zipfile as a BytesIO stream.
    Can take a filename or file-like object as an argument (including BytesIO or StringIO"""
    with ZipFileStreamer(fileobject) as zf:
        for filename in zf.namelist():
            yield filename, zf.extract_stream(filename)


class NemFile:
    def __init__(self, fileobject):
        self.__table_reader(fileobject)

    def __table_reader(self, fileobject):
        self.__dict__ = {}

        for line in fileobject.readlines():
            rows = line.decode().split(",")
            table = "{0}_{1}".format(rows[1], rows[2])

            # new table
            if rows[0] == "I":
                if table == "DISPATCH_OFFERTRK":
                    break
                self.__dict__[table] = line

            # append data to each table
            elif rows[0] == "D":
                self.__dict__[table] += line

        self.__dict__ = {
            table: pd.read_csv(BytesIO(self.__dict__[table]))
            for table in self.__dict__
        }

    def __iter__(self):
        for table in self.__dict__:
            yield table


def load_nemfile(f="PUBLIC_DISPATCHIS_201309150010_0000000246668929.zip"):
    with ZipFileStreamer(f) as zf:
        if zf.member_count == 1:
            filename = zf.namelist()[0]
            zfs = zf.extract_stream(filename)
            return NemFile(zfs)
        else:
            raise Exception("More than one file in zipfile")



import csv
import logging
import zipfile

from sqlalchemy.orm import sessionmaker

from opennem.db import db_connect
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class TableRecordSplitter(object):
    @check_spider_pipeline
    def process_item(self, item, spider):
        if not "tables" in item:
            logger.error(item)
            raise Exception("No tables passed to pipeline")

        tables = item["tables"]
        table = tables.pop()

        records = table["records"]

        for record in records:
            yield record


class UnzipSingleFilePipeline(object):
    @check_spider_pipeline
    def process_item(self, item, spider):

        if "body_stream" not in item:
            return item

        rs = item["body_stream"]
        content = ""

        with zipfile.ZipFile(rs) as zf:
            zip_files = zf.namelist()

            if len(zip_files) == 1:
                content = zf.open(zip_files[0])
                return {"file_handle": content, **item}

            if len(zip_files) != 1:
                raise Exception(
                    "Zero or more than one file in zip file. Have {}".format(
                        len(zip_files)
                    )
                )


class ReadStringHandle(object):
    @check_spider_pipeline
    def process_item(self, item, spider):
        if "file_handle" not in item:
            return item

        fh = item["file_handle"]

        content = fh.read()

        if type(content) is bytes:
            content = content.decode("utf-8")

        return {"content": content, **item}


class ExtractCSV(object):
    @check_spider_pipeline
    def process_item(self, item, spider):

        if "content" not in item:
            logger.error("No content in item to parse")
            return item

        content = item["content"]
        del item["content"]

        item["tables"] = {}
        table = {"name": None}

        content_split = content.splitlines()

        datacsv = csv.reader(content_split)

        for row in datacsv:
            if not row or type(row) is not list or len(row) < 1:
                continue

            record_type = row[0]

            if record_type == "C":
                # @TODO csv meta stored in table
                if table["name"] is not None:
                    table_name = table["name"]

                    if table_name in item["tables"]:
                        item["tables"][table_name]["records"] += table[
                            "records"
                        ]
                    else:
                        item["tables"][table_name] = table

            elif record_type == "I":
                if table["name"] is not None:
                    table_name = table["name"]

                    if table_name in item["tables"]:
                        item["tables"][table_name]["records"] += table[
                            "records"
                        ]
                    else:
                        item["tables"][table_name] = table

                table = {}
                table["name"] = "{}_{}".format(row[1], row[2])
                table["fields"] = fields = row[4:]
                table["records"] = []

            elif record_type == "D":
                values = row[4:]
                record = dict(zip(table["fields"], values))

                table["records"].append(record)

        return item


class DatabaseStore(object):
    def __init__(self):
        engine = db_connect()
        self.session = sessionmaker(bind=engine)



import csv
import logging
from io import StringIO
from typing import List

from sqlalchemy.sql.schema import Table

from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


def generate_csv_from_records(table: Table, records, column_names=None):
    if len(records) < 1:
        raise Exception("No records")

    csv_buffer = StringIO()

    table_column_names = [c.name for c in table.__table__.columns.values()]

    if isinstance(records, list):
        first_record = records[0]

        record_field_names = list(first_record.keys())

        for field_name in record_field_names:
            if field_name not in table_column_names:
                raise Exception("Column name not found: {}".format(field_name))

        for column_name in table_column_names:
            if column_name not in record_field_names:
                raise Exception(
                    "Missing value for column {}".format(column_name)
                )

        column_names = record_field_names

    if not column_names:
        column_names = table_column_names

    csvwriter = csv.DictWriter(csv_buffer, fieldnames=column_names)
    csvwriter.writeheader()

    # @TODO put the columns in the correct order ..

    for record in records:
        csvwriter.writerow(record)

    csv_buffer.seek(0)
    return csv_buffer


class RecordsToCSVPipeline(object):
    @check_spider_pipeline
    def process_item(self, item: List[dict], spider=None):
        if not isinstance(item, list):
            item = [item]

        for record_set in item:
            if not isinstance(record_set, dict):
                logger.error(
                    "Invalid record_set passed to CSV pipeline: %s", record_set
                )
                return record_set

            if "table_schema" not in record_set:
                logger.error("No table model passed to csv generator")
                logger.error(record_set)
                return record_set

            table = record_set["table_schema"]

            if "records" not in record_set:
                logger.error("No records to generate csv from")
                return record_set

            records = record_set["records"]

            csv_content = generate_csv_from_records(table, records)

            record_set["csv"] = csv_content

        return item



import json
import os

from opennem.core.loader import load_data

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")

FACILITY_STATIONS = load_data("facility_stations.json")


def facility_station_join_by_name(station_name):
    """
        This is a bit of a hack where we join facilities to stations by name only
        where there is no duid or other data we can join on.

        Used a lot in updating AEMO data.

        The list of station names to do this based on is stored in data/facility_stations.json

    """

    if not type(FACILITY_STATIONS) is list:
        raise Exception("Error loading facility station map: not a list")

    return station_name not in FACILITY_STATIONS



import logging
import re
from datetime import datetime
from pathlib import Path
from typing import Optional
from urllib.parse import ParseResult, urlparse

import scrapy
from scrapy import Spider

from opennem.schema.network import NetworkNEM
from opennem.utils.dates import parse_date

PADDING_WIDTH = 3

__is_number = re.compile(r"^\d+$")


def is_number(value):
    if re.match(__is_number, value):
        return True
    return False


def parse_dirlisting(raw_string):
    """
        given a raw text directory listing like "     Saturday 11th June 2020      6789"
        will parse and return both the date and listing type

        @param raw_string - the raw directory listing string
        @return dict of the date in iso format and the type (file or directory)
    """
    components = raw_string.split(" " * PADDING_WIDTH)
    components = [i.strip() for i in components]
    components = list(filter(lambda x: x != "", components))

    _ltype = "dir"

    if not components or len(components) < 2:
        logging.debug(components)
        raise Exception(
            "Invalid line string: {}. Components are: {}".format(
                raw_string, components
            )
        )

    if is_number(components[1]):
        _ltype = "file"

    dt = parse_date(components[0], network=NetworkNEM)

    if type(dt) is not datetime:
        raise Exception(
            "{} is not a valid datetime. Original value was {}".format(
                dt, components[0]
            )
        )

    return {
        "date": dt.isoformat(),
        "type": _ltype,
    }


def get_file_extensions(url: str) -> Optional[str]:
    url_parsed: ParseResult = urlparse(url)

    if not url_parsed.path:
        return None

    url_path = Path(url_parsed.path)

    return url_path.suffix


class DirlistingSpider(Spider):
    """
        spider that parses html directory listings produced by web servers


    """

    limit = 0

    skip = 0

    start_url = None

    supported_extensions = [".csv", ".zip"]

    def start_requests(self):
        if self.custom_settings is None:
            self.custom_settings = {}

        starts = None

        if (
            hasattr(self, "start_url")
            and self.start_url
            and type(self.start_url) is str
        ):
            starts = [self.start_url]

        if (
            hasattr(self, "start_urls")
            and self.start_urls
            and type(self.start_urls) is list
        ):
            starts = [self.start_urls]

        for url in starts:
            yield scrapy.Request(url)

    def parse(self, response):
        links = list(
            reversed(
                [
                    i.get()
                    for i in response.xpath(
                        "//body/pre/br/following-sibling::a/@href"
                    )
                ]
            )
        )

        metadata = list(
            reversed(
                [
                    parse_dirlisting(i.get())
                    for i in response.xpath(
                        "//body/pre/br/following-sibling::text()"
                    )
                ]
            )
        )

        parsed = 0

        if self.limit > 0 and self.skip > 0:
            self.limit = self.limit + self.skip

        for i, entry in enumerate(metadata):
            if entry["type"] == "file":
                link = response.urljoin(links[i])

                if self.limit and self.limit > 0 and (parsed >= self.limit):
                    self.log(f"Reached limit of {self.limit}", logging.INFO)
                    return None

                if get_file_extensions(link) not in self.supported_extensions:
                    self.log(
                        "Not a supported extension: {} {}".format(
                            link[-4:].lower(),
                            ", ".join(self.supported_extensions),
                        ),
                        logging.INFO,
                    )
                    continue

                parsed += 1

                if self.skip and self.skip >= parsed:
                    self.log(
                        f"Skipping entry {link}", logging.DEBUG,
                    )
                    continue

                self.log("Getting {}".format(link), logging.INFO)

                yield {"link": link, **entry}




"""FIX: cast m nulls to 0 in om_energy_sum

Revision ID: f87394cc4c00
Revises: 4fb955a7a2a7
Create Date: 2020-10-21 18:44:31.318648

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "f87394cc4c00"
down_revision = "4fb955a7a2a7"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
    create or replace function onenergy_finalfunc(agg_state tenergy)
    returns numeric
    immutable
    language plpgsql
    as $$
    declare
        result_int numeric;
        num_intervals int;
        poly text = '';
        poly_array text[] = '{}';
        interval_size_minutes int;
        y int := 0;
        m int;
    begin
        num_intervals := array_length(agg_state.items, 1);
        interval_size_minutes := agg_state.bucket_size_minutes / num_intervals;

        poly_array = poly_array || format('%s %s', 0, 0);

        FOREACH m IN ARRAY agg_state.items
        LOOP
            m := coalesce(m, 0);

            poly_array = poly_array || format('%s %s', y, m);
            y := y + interval_size_minutes;

        END LOOP;

        poly_array = poly_array || format('%s %s', agg_state.bucket_size_minutes, 0);
        poly_array = poly_array || format('%s %s', 0, 0);


        poly := format('POLYGON((%s))', array_to_string(poly_array, ', '));

        result_int := (ST_AREA(poly::geometry) / 60)::numeric;

        return result_int;
    end;
    $$;
    """
    )


def downgrade():
    pass



from datetime import datetime
from typing import List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session
from starlette import status

from opennem.db import get_database_session
from opennem.db.models.opennem import Facility, Location, Revision, Station

from .schema import (
    RevisionModification,
    RevisionModificationResponse,
    UpdateResponse,
)

router = APIRouter()


@router.get("/")
def revisions(
    session: Session = Depends(get_database_session),
) -> List[Revision]:
    revisions = session.query(Revision).all()

    return revisions


@router.get("/{revision_id}")
def revision(
    revision_id: int, session: Session = Depends(get_database_session)
):
    revision = session.query(Revision).get(revision_id)

    if not revision:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Revision not found"
        )

    return revision


REVISION_TARGET_MAP = {
    "station": Station,
    "facility": Facility,
    "location": Location,
}


@router.put("/{revision_id}", name="revision update")
def revision_update(
    data: RevisionModification = {},
    session: Session = Depends(get_database_session),
    revision_id: int = None,
) -> dict:
    revision: Revision = session.query(Revision).get(revision_id)

    if not revision:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Revision not found"
        )

    if data.modification == "reject":
        revision.discarded = True
        revision.approved = False
        revision.discarded_by = "opennem.admin"
        revision.discarded_at = datetime.now()

        session.add(revision)
        session.commit()

        response = UpdateResponse(success=True, record=revision)

    revision.approved = True
    revision.approved_at = datetime.now()
    revision.approved_by = "opennem.admin"

    if revision.parent_type not in REVISION_TARGET_MAP.keys():
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail="Revision schema type not supported",
        )

    revision_target_type = REVISION_TARGET_MAP[revision.parent_type]

    revision_target = session.query(revision_target_type).get(
        revision.parent_id
    )

    if not revision_target:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Revision target record not found",
        )

    for target_field, target_value in revision.changes.items():
        if hasattr(revision_target, f"{target_field}_code"):
            target_field = "{}_code".format(target_field)

        if hasattr(revision_target, f"{target_field}_id"):
            target_field = "{}_id".format(target_field)

        if not hasattr(revision_target, target_field):
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Invalid revision field",
            )

        try:
            setattr(revision_target, target_field, target_value)
        except Exception:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Error setting revision",
            )

    session.add(revision_target)
    session.add(revision)
    session.commit()

    response = RevisionModificationResponse(
        success=True, record=revision, target=revision_target
    )

    return response


@router.post("/approve/{revision_id}", name="revision approve")
def revision_approve(
    data: RevisionModification = {},
    session: Session = Depends(get_database_session),
    revision_id: int = None,
) -> dict:
    revision = session.query(Revision).get(revision_id)

    if not revision:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Revision not found"
        )

    revision.approved = True
    revision.approved_by = "opennem.admin"

    session.add(revision)
    session.commit()

    response = UpdateResponse(success=True, record=revision)

    return response


@router.post("/reject/{revision_id}", name="revision reject")
def revision_reject(
    data: RevisionModification = {},
    session: Session = Depends(get_database_session),
    revision_id: int = None,
) -> dict:
    revision: Revision = session.query(Revision).get(revision_id)

    if not revision:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Revision not found"
        )

    revision.approved = False
    revision.discarded = False
    revision.discarded_by = "opennem.admin"
    revision.discarded_at = datetime.now()

    session.add(revision)
    session.commit()

    response = UpdateResponse(success=True, record=revision)

    return response




import scrapy

from opennem.spiders.bom.bom_base import BomJSONObservationSpider

from .utils import get_stations, get_stations_priority


class BomCapitalsSpider(BomJSONObservationSpider):
    name = "bom.capitals"

    def start_requests(self):
        priority_stations = get_stations_priority()

        for station in priority_stations:
            yield scrapy.Request(station.feed_url, meta={"code": station.code})


class BomAllSpider(BomJSONObservationSpider):
    name = "bom.all"

    def start_requests(self):
        station = get_stations()

        for station in station:
            yield scrapy.Request(station.feed_url, meta={"code": station.code})



"""
    settings files - read settings from env


    will read in order:

        * .env
        * .env.{environment}
        * system env
        * pydantic settings

"""


import os
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv

from opennem.settings.utils import load_env_file

from .schema import OpennemSettings

if Path(".env").is_file():
    load_dotenv()

MODULE_DIR = os.path.dirname(__file__)

ENV = os.getenv("ENV", default="development")

# settings default loader

settings: Optional[OpennemSettings] = None

_env_file = load_env_file(ENV)

# @TODO add logging
if _env_file:
    load_dotenv(dotenv_path=_env_file, override=True)

# @NOTE don't use pydantics env file support since it doesn't support multiple
settings = OpennemSettings()



from typing import List

import pint

from opennem.core.loader import load_data
from opennem.schema.units import UnitDefinition

unit_registry = pint.UnitRegistry()
Q = unit_registry.Quantity


def load_units() -> List[UnitDefinition]:
    units_dics = load_data("units.json")

    units = [UnitDefinition(**i) for i in units_dics]

    # unique aliases in set
    assert len([i.name_alias for i in units if i.name_alias]) == len(
        set([i.name_alias for i in units if i.name_alias])
    ), "Unique names for aliases required"

    # unique names in
    assert len([i.name for i in units]) == len(
        set([i.name for i in units])
    ), "Unique unit names required"

    return units


UNITS = load_units()

UNITS_SUPPORTED = [i.unit for i in UNITS]

UNIT_TYPES_SUPPORTED = set([i.unit_type for i in UNITS])


def get_unit(unit_human: str) -> UnitDefinition:
    global UNITS

    # Lazy load
    if not UNITS:
        UNITS = load_units()

    unit_lookup = list(
        filter(
            lambda x: x.name_alias == unit_human or x.name == unit_human, UNITS
        )
    )

    if unit_lookup:
        return unit_lookup.pop()

    raise Exception("Invalid interval {} not mapped".format(unit_human))



from typing import Optional

from opennem.api.schema import ApiBase


class Photo(ApiBase):
    id: int

    width: int
    height: int
    photo_url: Optional[str]

    license_type: Optional[str]
    license_link: Optional[str]
    author: Optional[str]
    author_link: Optional[str]



import csv
import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.db.models.opennem import BalancingSummary
from opennem.pipelines import DatabaseStoreBase
from opennem.schema.network import NetworkWEM
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class WemStoreBalancingSummary(DatabaseStoreBase):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = self.session()

        csvreader = csv.DictReader(item["content"].split("\n"))

        records_to_store = []

        for record in csvreader:
            trading_interval = parse_date(
                record["Trading Interval"], network=NetworkWEM
            )

            if not trading_interval:
                continue

            records_to_store.append(
                {
                    "created_by": spider.name,
                    "network_id": "WEM",
                    "network_region": "WEM",
                    "trading_interval": trading_interval,
                    "forecast_load": record["Load Forecast (MW)"],
                    "generation_scheduled": record[
                        "Scheduled Generation (MW)"
                    ],
                    "generation_non_scheduled": record[
                        "Non-Scheduled Generation (MW)"
                    ],
                    "generation_total": record["Total Generation (MW)"],
                    "price": record["Final Price ($/MWh)"],
                }
            )

        stmt = insert(BalancingSummary).values(records_to_store)
        stmt.bind = self.engine
        stmt = stmt.on_conflict_do_update(
            index_elements=[
                "trading_interval",
                "network_id",
                "network_region",
            ],
            set_={
                "price": stmt.excluded.price,
                "generation_total": stmt.excluded.generation_total,
            },
        )

        try:
            s.execute(stmt)
            s.commit()
        except Exception as e:
            logger.error("Error inserting records")
            logger.error(e)
            return 0
        finally:
            s.close()

        return len(records_to_store)



"""
    Sets up logging for the main opennem project


"""

import logging
from logging import config

from opennem.settings.logging import get_logging_config

logging_config = get_logging_config()
config.dictConfig(logging_config)



import logging
from datetime import date, datetime, timedelta, timezone
from typing import List, Optional, Union

from dateutil.parser import ParserError, parse

from opennem.core.networks import NetworkSchema
from opennem.core.normalizers import normalize_whitespace
from opennem.utils.timezone import UTC, is_aware, make_aware

logger = logging.getLogger(__name__)

# Date formats
# See: https://docs.python.org/3.8/library/datetime.html#strftime-and-strptime-behavior
DATE_FORMATS = [
    "%Y/%m/%d,%H:%M:%S",
    "%Y%m%d",
    #
    "%d/%m/%y %H:%M",
    "%Y/%m/%d %H:%M:%S",
    "%d/%m/%Y %I:%M:%S %p",  # excel format
    "%Y%m%d%H%M%S",  # bom format
]


def optimized_data_parser(date_str: str) -> datetime:
    """
        Turns out that dateutil's date parser is slow since
        it does a lot of string parsing. Here we try matching
        the date using known string formats.
    """
    dt_return = None

    date_str = normalize_whitespace(date_str).strip()

    for date_format_str in DATE_FORMATS:
        try:
            dt_return = datetime.strptime(date_str, date_format_str)
        except (KeyError, OverflowError, ValueError, AttributeError):
            continue
        if dt_return:
            return dt_return

    return dt_return


def parse_date(
    date_str: Union[str, datetime],
    network: Optional[NetworkSchema] = None,
    dayfirst: bool = True,
    yearfirst: bool = False,
    is_utc: bool = False,
    timezone: timezone = False,
    use_optimized: bool = True,
) -> Optional[datetime]:
    dt_return = None

    if isinstance(date_str, datetime):
        dt_return = date_str

    elif isinstance(date_str, str):
        if use_optimized:
            dt_return = optimized_data_parser(date_str)

        if not dt_return:
            try:
                dt_return = parse(
                    date_str, dayfirst=dayfirst, yearfirst=yearfirst
                )
            except ParserError:
                raise ValueError("Invalid date string passed")

    else:
        raise ValueError("Require a datetime or string object to parse date")

    if network:
        tz = network.get_timezone()

        if tz:
            if is_aware(dt_return):
                if hasattr(tz, "localize"):
                    dt_return = tz.localize()
                else:
                    dt_return = dt_return.replace(tzinfo=tz)
            else:
                dt_return = make_aware(dt_return, timezone=tz)

    if is_utc:
        tz = UTC

        if is_aware(dt_return):
            if hasattr(tz, "localize"):
                dt_return = tz.localize()
            else:
                dt_return = dt_return.replace(tzinfo=tz)
        else:
            dt_return = make_aware(dt_return, timezone=tz)

    if timezone:
        dt_return = make_aware(dt_return, timezone=timezone)

    return dt_return


def get_date_component(format_str: str) -> str:
    """
        Get a part of a date
    """

    return datetime.now().strftime(format_str)


def date_series(
    start: Union[datetime, date] = None,
    end: Union[datetime, date] = None,
    length: int = 30,
    interval: timedelta = timedelta(days=1),
    reverse: bool = False,
) -> List[datetime]:
    """
        Generate a datetime series

    """
    if start and isinstance(start, datetime):
        start = start.date()

    if end and isinstance(end, datetime):
        end = start.date()

    if not start:
        start = datetime.now().date()

    if end:
        length = int(abs((end - start) / interval))

    next_record = start

    for _ in range(length):
        yield next_record

        if reverse:
            next_record -= interval
        else:
            next_record += interval




from typing import List, Optional

from opennem.api.export.queries import (
    energy_network_fueltech_query,
    power_network_fueltech_query,
    price_network_query,
)
from opennem.api.stats.controllers import get_scada_range, stats_factory
from opennem.api.stats.schema import DataQueryResult, OpennemDataSet
from opennem.api.time import human_to_interval, human_to_period
from opennem.api.weather.queries import (
    observation_query,
    observation_query_all,
)
from opennem.core.networks import network_from_network_code
from opennem.core.units import get_unit
from opennem.db import get_database_engine
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimePeriod


def weather_daily(
    network_code: str,
    station_code: str,
    year: Optional[int] = None,
    unit_name: str = "temperature_mean",
    period_human: str = None,
    include_min_max: bool = True,
):
    station_codes = []

    if station_code:
        station_codes = [station_code]

    engine = get_database_engine()
    interval = human_to_interval("1d")
    network = network_from_network_code(network_code)
    units = get_unit(unit_name)
    scada_range = get_scada_range(network=network)

    if year:
        query = observation_query(
            station_codes=station_codes,
            interval=interval,
            network=network,
            scada_range=scada_range,
            year=year,
        )
    elif period_human:
        interval = human_to_interval("{}m".format(network.interval_size))
        period = human_to_period("7d")

        query = observation_query(
            station_codes=station_codes,
            interval=interval,
            network=network,
            scada_range=scada_range,
            period=period,
        )
    else:
        interval = human_to_interval("1M")

        query = observation_query_all(
            station_codes=station_codes,
            scada_range=scada_range,
            network=network,
        )

    with engine.connect() as c:
        row = list(c.execute(query))

    temp_avg = [
        DataQueryResult(
            interval=i[0], group_by=i[1], result=i[2] if len(i) > 1 else None
        )
        for i in row
    ]

    temp_min = [
        DataQueryResult(
            interval=i[0], group_by=i[1], result=i[3] if len(i) > 1 else None
        )
        for i in row
    ]

    temp_max = [
        DataQueryResult(
            interval=i[0], group_by=i[1], result=i[4] if len(i) > 1 else None
        )
        for i in row
    ]

    if len(temp_avg) < 1:
        raise Exception("No results from query: {}".format(query))

    stats = stats_factory(
        stats=temp_avg,
        units=units,
        network=network,
        interval=interval,
        code="bom",
        group_field="temperature",
    )

    if include_min_max:
        stats_min = stats_factory(
            stats=temp_min,
            units=get_unit("temperature_min"),
            network=network,
            interval=interval,
            code="bom",
            group_field="temperature",
        )

        stats_max = stats_factory(
            stats=temp_max,
            units=get_unit("temperature_max"),
            network=network,
            interval=interval,
            code="bom",
            group_field="temperature",
        )

        stats.append_set(stats_min)
        stats.append_set(stats_max)

    return stats


def power_week(
    network_code: str = "WEM",
    network_region_code: str = None,
    networks: Optional[List[NetworkSchema]] = None,
) -> OpennemDataSet:
    engine = get_database_engine()

    network = network_from_network_code(network_code)
    interval = human_to_interval("{}m".format(network.interval_size))
    period = human_to_period("7d")
    units = get_unit("power")
    network_region = None

    query = power_network_fueltech_query(
        network=network,
        networks=networks,
        period=period,
        interval=interval,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        row = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None
        )
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    result = stats_factory(
        stats,
        code=network_region_code or network.code,
        network=network,
        interval=interval,
        period=period,
        units=units,
        region=network_region,
        fueltech_group=True,
    )

    if not result:
        raise Exception("No results")

    # price

    query = price_network_query(
        network=network,
        networks=networks,
        period=period,
        interval=interval,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        row = list(c.execute(query))

    stats_price = [
        DataQueryResult(
            interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None
        )
        for i in row
    ]

    stats_market_value = stats_factory(
        stats=stats_price,
        code=network_region_code or network.code.lower(),
        units=get_unit("price_energy_mega"),
        network=network,
        interval=interval,
        region=network_region_code,
        period=period,
    )

    result.append_set(stats_market_value)

    return result


def energy_fueltech_daily(
    network: NetworkSchema,
    year: Optional[int] = None,
    network_region_code: Optional[str] = None,
    interval_size: str = "1d",
    networks: Optional[List[NetworkSchema]] = None,
) -> OpennemDataSet:
    engine = get_database_engine()
    period: TimePeriod = human_to_period("1Y")
    interval = human_to_interval(interval_size)
    units = get_unit("energy_giga")

    query = energy_network_fueltech_query(
        year=year,
        network=network,
        network_region=network_region_code,
        networks=networks,
    )

    with engine.connect() as c:
        row = list(c.execute(query))

    results_energy = [
        DataQueryResult(
            interval=i[0], group_by=i[1], result=i[2] if len(i) > 1 else None
        )
        for i in row
    ]

    results_market_value = [
        DataQueryResult(
            interval=i[0], group_by=i[1], result=i[3] if len(i) > 1 else None
        )
        for i in row
    ]

    if len(results_energy) < 1:
        raise Exception("No results from query: {}".format(query))

    stats = stats_factory(
        stats=results_energy,
        units=units,
        network=network,
        fueltech_group=True,
        interval=interval,
        region=network.code.lower(),
        period=period,
        code=network.code.lower(),
    )

    stats_market_value = stats_factory(
        stats=results_market_value,
        units=get_unit("market_value"),
        network=network,
        fueltech_group=True,
        interval=interval,
        region=network.code.lower(),
        period=period,
        code=network.code.lower(),
    )

    stats.append_set(stats_market_value)

    return stats




"""observation_time timezone

Revision ID: 637faac3f9c1
Revises: f87394cc4c00
Create Date: 2020-10-22 00:43:25.578149

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "637faac3f9c1"
down_revision = "f87394cc4c00"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "ALTER TABLE bom_observation ALTER COLUMN observation_time TYPE TIMESTAMP WITH TIME ZONE USING observation_time AT TIME ZONE 'UTC'"
    )


def downgrade():
    op.execute(
        "ALTER TABLE bom_observation ALTER COLUMN observation_time TYPE TIMESTAMP WITHOUT TIME ZONE"
    )




from opennem.importer.aemo_gi import gi_export
from opennem.importer.aemo_rel import rel_export
from opennem.importer.mms import mms_export
from opennem.importer.registry import registry_export


def run_all():
    mms_export()
    rel_export()
    gi_export()
    registry_export()



from huey import RedisHuey, crontab

from opennem.api.export.map import PriorityType
from opennem.api.export.tasks import (
    export_energy,
    export_metadata,
    export_power,
    wem_export_daily,
    wem_export_monthly,
    wem_export_power,
)
from opennem.exporter.geojson import export_facility_geojson
from opennem.monitors.aemo_intervals import get_wem_interval_delay
from opennem.settings import settings

huey = RedisHuey("opennem.exporter", host=settings.cache_url.host)

# export tasks


@huey.periodic_task(crontab(minute="*/5",))
def schedule_wem_export_task():
    wem_export_power()
    export_power(PriorityType.live)


@huey.periodic_task(crontab(hour="*/1"))
def schedule_wem_export_daily_most_recent():
    wem_export_daily(limit=1)
    export_energy(PriorityType.daily, latest=True)


@huey.periodic_task(crontab(hour="*/12"))
def schedule_wem_export_years():
    wem_export_daily()
    export_energy(PriorityType.daily)


@huey.periodic_task(crontab(hour="*/12"))
def schedule_wem_export_all():
    wem_export_monthly()
    export_energy(PriorityType.monthly)


@huey.periodic_task(crontab(minute="*/30"))
def schedule_export_geojson():
    export_facility_geojson()


@huey.periodic_task(crontab(minute="*/30"))
def schedule_export_au():
    export_metadata()


# monitoring tasks
@huey.periodic_task(crontab(minute="*/5"))
def monitor_wem_interval():
    get_wem_interval_delay()



from datetime import datetime
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session
from starlette import status

from opennem.api.time import human_to_interval, human_to_period
from opennem.core.networks import network_from_network_code
from opennem.core.normalizers import normalize_duid
from opennem.core.units import get_unit
from opennem.db import get_database_engine, get_database_session
from opennem.db.models.opennem import Facility, Station
from opennem.schema.time import TimePeriod
from opennem.utils.time import human_to_timedelta

from .controllers import get_scada_range, stats_factory
from .queries import (
    energy_facility,
    energy_network,
    energy_network_fueltech,
    energy_network_fueltech_all,
    energy_network_fueltech_year,
    power_facility,
    power_network_fueltech,
    price_network_monthly,
    price_network_region,
)
from .schema import DataQueryResult, OpennemDataSet

router = APIRouter()


@router.get(
    "/power/unit/{network_code}/{unit_code:path}",
    name="stats:Unit Power",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def power_unit(
    unit_code: str = Query(..., description="Unit code"),
    network_code: str = Query(..., description="Network code"),
    since: str = Query(None, description="Since as human interval"),
    since_dt: datetime = Query(None, description="Since as datetime"),
    interval: str = Query(None, description="Interval"),
    period: str = Query("7d", description="Period"),
    session: Session = Depends(get_database_session),
    engine=Depends(get_database_engine),
) -> OpennemDataSet:
    if not since:
        since = datetime.now() - human_to_timedelta("7d")

    network = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No such network",
        )

    if not interval:
        interval = "{}m".format(network.interval_size)

    interval = human_to_interval(interval)
    period = human_to_period(period)
    units = get_unit("power")

    stats = []

    facility_codes = [normalize_duid(unit_code)]

    query = power_facility(
        facility_codes, network.code, interval=interval, period=period
    )

    with engine.connect() as c:
        results = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None
        )
        for i in results
    ]

    if len(stats) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Unit stats not found",
        )

    output = stats_factory(
        stats,
        code=unit_code,
        interval=interval,
        period=period,
        units=units,
        network=network,
    )

    return output


@router.get(
    "/power/station/{network_code}/{station_code:path}",
    name="stats:Station Power",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def power_station(
    station_code: str = Query(..., description="Station code"),
    network_code: str = Query(..., description="Network code"),
    since: datetime = Query(None, description="Since time"),
    interval: str = Query(None, description="Interval"),
    period: str = Query("7d", description="Period"),
    session: Session = Depends(get_database_session),
    engine=Depends(get_database_engine),
) -> OpennemDataSet:
    if not since:
        since = datetime.now() - human_to_timedelta("7d")

    network = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No such network",
        )

    if not interval:
        # @NOTE rooftop data is 15m
        if station_code.startswith("ROOFTOP"):
            interval = "15m"
        else:
            interval = "{}m".format(network.interval_size)

    interval = human_to_interval(interval)
    period = human_to_period(period)
    units = get_unit("power")

    station = (
        session.query(Station)
        .join(Facility)
        .filter(Station.code == station_code)
        .filter(Facility.network_id == network.code)
        .filter(Station.approved == True)
        .one_or_none()
    )

    if not station:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Station not found"
        )

    facility_codes = list(set([f.code for f in station.facilities]))

    stats = []

    query = power_facility(
        facility_codes, network_code, interval=interval, period=period
    )

    # print(query)

    with engine.connect() as c:
        results = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None
        )
        for i in results
    ]

    if len(stats) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station stats not found",
        )

    result = stats_factory(
        stats,
        code=station_code,
        network=network,
        interval=interval,
        period=period,
        units=units,
    )

    return result


@router.get(
    "/power/fueltech/{network_code}/{station_code:path}",
    name="stats:Network Fueltech Power",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def power_network_fueltech_api(
    network_code: str = Query(..., description="Network code"),
    network_region: str = Query(None, description="Network region"),
    interval: str = Query(None, description="Interval"),
    period: str = Query("7d", description="Period"),
    engine=Depends(get_database_engine),
) -> OpennemDataSet:
    network = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No such network",
        )

    if not interval:
        interval = "{}m".format(network.interval_size)

    interval = human_to_interval(interval)
    period = human_to_period(period)
    units = get_unit("power")

    scada_range = get_scada_range(network=network)

    query = power_network_fueltech(
        network=network,
        interval=interval,
        period=period,
        network_region=network_region,
        scada_range=scada_range,
    )

    with engine.connect() as c:
        results = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None
        )
        for i in results
    ]

    if len(stats) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station stats not found",
        )

    result = stats_factory(
        stats,
        code=network.code,
        network=network,
        interval=interval,
        period=period,
        units=units,
        region=network_region,
        fueltech_group=True,
    )

    return result


"""
    Energy endpoints
"""


@router.get(
    "/energy/station/{network_code}/{station_code}",
    name="Energy Station",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def energy_station(
    engine=Depends(get_database_engine),
    session: Session = Depends(get_database_session),
    network_code: str = Query(..., description="Network code"),
    station_code: str = Query(..., description="Station Code"),
    interval: str = Query(None, description="Interval"),
    period: str = Query("7d", description="Period"),
) -> OpennemDataSet:
    """
        Get energy output for a station (list of facilities)
        over a period
    """

    network = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No such network",
        )

    if not interval:
        # @NOTE rooftop data is 15m
        if station_code.startswith("ROOFTOP"):
            interval = "15m"
        else:
            interval = "{}m".format(network.interval_size)

    interval = human_to_interval(interval)
    period = human_to_period(period)
    units = get_unit("energy")

    station = (
        session.query(Station)
        .join(Station.facilities)
        .filter(Station.code == station_code)
        .filter(Facility.network_id == network.code)
        .one_or_none()
    )

    if not station:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Station not found"
        )

    if len(station.facilities) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station has no facilities",
        )

    facility_codes = list(set([f.code for f in station.facilities]))

    query = energy_facility(
        facility_codes, network_code, interval=interval, period=period
    )

    with engine.connect() as c:
        results = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None
        )
        for i in results
    ]

    if len(stats) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station stats not found",
        )

    result = stats_factory(
        stats,
        code=station_code,
        network=network,
        interval=interval,
        period=period,
        units=units,
    )

    return result


@router.get(
    "/energy/network/{network_code}",
    name="Energy Network",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def energy_network_api(
    engine=Depends(get_database_engine),
    network_code: str = Query(..., description="Network code"),
    interval: str = Query("1d", description="Interval"),
    period: str = Query("1Y", description="Period"),
) -> OpennemDataSet:

    results = []

    network = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No such network",
        )

    if not interval:
        interval = "{}m".format(network.interval_size)

    interval = human_to_interval(interval)
    period = human_to_period(period)
    units = get_unit("energy_giga")

    query = energy_network(network=network, interval=interval, period=period)

    with engine.connect() as c:
        results = list(c.execute(query))

    if len(results) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No results"
        )

    stats = [
        DataQueryResult(
            interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None
        )
        for i in results
    ]

    result = stats_factory(
        stats,
        code=network.code,
        network=network,
        interval=interval,
        period=period,
        units=units,
    )

    return result


@router.get(
    "/energy/fueltech/{network_code}/{station_code:path}",
    name="stats:Network Fueltech Energy",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def energy_network_fueltech_api(
    network_code: str = Query(None, description="Network code"),
    network_region: str = Query(None, description="Network region"),
    interval: str = Query("1d", description="Interval"),
    year: int = Query(None, description="Year to query"),
    period: str = Query("1Y", description="Period"),
    engine=Depends(get_database_engine),
) -> OpennemDataSet:
    network = network_from_network_code(network_code)
    interval = human_to_interval(interval)

    period_obj = None

    if period:
        period_obj: TimePeriod = human_to_period(period)

    units = get_unit("energy_giga")

    query = ""

    if year and isinstance(year, int):
        period = human_to_period("1Y")

        if year > datetime.now().year or year < 1996:
            raise HTTPException(
                status_code=status.HTTP_406_NOT_ACCEPTABLE,
                detail="Not a valid year",
            )

        scada_range = get_scada_range(network=network)

        query = energy_network_fueltech_year(
            network=network,
            interval=interval,
            year=year,
            network_region=network_region,
            scada_range=scada_range,
        )
    elif period_obj and period_obj.period_human == "all":
        scada_range = get_scada_range(network=network)

        query = energy_network_fueltech_all(
            network=network,
            network_region=network_region,
            scada_range=scada_range,
        )
    else:
        query = energy_network_fueltech(
            network=network,
            interval=interval,
            period=period_obj,
            network_region=network_region,
        )

    # print(query)

    with engine.connect() as c:
        results = list(c.execute(query))

    stats = [
        DataQueryResult(
            interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None
        )
        for i in results
    ]

    if len(stats) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Energy stats not found",
        )

    result = stats_factory(
        stats,
        code=network.code,
        network=network,
        interval=interval,
        period=period_obj,
        units=units,
        region=network_region,
        fueltech_group=True,
    )

    if not result or len(result.data) == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No stats"
        )

    return result


"""
    Price endpoints
"""


@router.get(
    "/price/{network_code}/{network_region_code}",
    name="Price Network Region",
    response_model=OpennemDataSet,
    response_model_exclude_unset=True,
)
def price_network_region_api(
    engine=Depends(get_database_engine),
    network_code: str = Query(..., description="Network code"),
    network_region_code: str = Query(..., description="Region code"),
    interval: str = Query(None, description="Interval"),
    period: str = Query("7d", description="Period"),
    year: Optional[int] = None,
) -> OpennemDataSet:
    network = network_from_network_code(network_code)

    if not network:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No such network",
        )

    if not interval:
        interval = "{}m".format(network.interval_size)

    interval = human_to_interval(interval)

    period_obj = None

    if period:
        period_obj = human_to_period(period)

    units = get_unit("price")

    scada_range = get_scada_range(network=network)

    if (
        period_obj
        and period_obj.period_human == "all"
        and interval.interval_human == "1M"
    ):
        query = price_network_monthly(
            network=network,
            network_region_code=network_region_code,
            scada_range=scada_range,
        )
    else:
        query = price_network_region(
            network=network,
            network_region_code=network_region_code,
            interval=interval,
            period=period_obj,
            scada_range=scada_range,
            year=year,
        )

    with engine.connect() as c:
        results = list(c.execute(query))

    if len(results) < 1:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="No data found"
        )

    stats = [
        DataQueryResult(
            interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None
        )
        for i in results
    ]

    result = stats_factory(
        stats,
        code=network.code,
        region=network_region_code,
        network=network,
        interval=interval,
        period=period_obj,
        units=units,
        group_field="price",
    )

    return result



from opennem.core.loader import load_data

STATION_NAME_MAP = load_data("station_name_maps.json")


def station_map_name(station_name):
    """
        Clean names manually using a map

    """
    if type(station_name) is not str:
        return station_name

    if not type(STATION_NAME_MAP) is dict:
        raise Exception("Error loading station name maps: not a dict")

    if station_name in STATION_NAME_MAP:
        return STATION_NAME_MAP[station_name]

    for map_name in STATION_NAME_MAP.keys():
        if station_name.lower().startswith(map_name.lower()):
            return STATION_NAME_MAP[map_name]

    return station_name



import logging
from datetime import datetime
from typing import List, Optional

from opennem.api.export.controllers import (
    energy_fueltech_daily,
    power_week,
    weather_daily,
)
from opennem.api.export.map import (
    PriorityType,
    StatExport,
    StatType,
    get_export_map,
)
from opennem.api.export.utils import write_output
from opennem.api.stats.router import price_network_region_api
from opennem.db import get_database_engine
from opennem.schema.network import NetworkWEM
from opennem.settings import settings

# @TODO q this ..
# @NOTE done in new stat export maps
YEAR_MIN = 2010

logger = logging.getLogger(__name__)

export_map = get_export_map()


def export_power(priority: Optional[PriorityType] = None):
    power_stat_sets = export_map.get_by_stat_type(StatType.power, priority)

    for power_stat in power_stat_sets:
        stat_set = power_week(
            network_code=power_stat.network.code,
            network_region_code=power_stat.network_region,
            networks=power_stat.networks,
        )

        if power_stat.bom_station:
            weather_set = weather_daily(
                station_code=power_stat.bom_station,
                network_code=power_stat.network.code,
                include_min_max=False,
                period_human="7d",
                unit_name="temperature",
            )

            stat_set.append_set(weather_set)

        write_output(power_stat.path, stat_set)


def export_energy(
    priority: Optional[PriorityType] = None, latest: bool = False
):
    energy_stat_sets = export_map.get_by_stat_type(StatType.energy, priority)

    CURRENT_YEAR = datetime.now().year

    for energy_stat in energy_stat_sets:
        if energy_stat.year:

            if latest and energy_stat.year != CURRENT_YEAR:
                continue

            stat_set = energy_fueltech_daily(
                year=energy_stat.year,
                network=energy_stat.network,
                networks=energy_stat.networks,
                network_region_code=energy_stat.network_region,
            )

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)

        elif energy_stat.period.period_human == "all":
            stat_set = energy_fueltech_daily(
                interval_size="1M",
                network=energy_stat.network,
                networks=energy_stat.networks,
                network_region_code=energy_stat.network_region,
            )

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)


def export_metadata():

    _export_map_out = export_map

    # this is a hack because pydantic doesn't
    # serialize properties
    for r in _export_map_out.resources:
        r.file_path = r.path()

    write_output("metadata.json", _export_map_out)


def wem_export_power(is_local: bool = False):
    engine = get_database_engine()

    stat_set = power_week(network_code="WEM")

    stat_set.data += weather_daily(
        station_code="009021",
        network_code="WEM",
        include_min_max=False,
        period_human="7d",
        unit_name="temperature",
    ).data

    price = price_network_region_api(
        engine=engine,
        network_code="WEM",
        network_region_code="WEM",
        interval="30m",
        period="7d",
    )

    stat_set.data = stat_set.data + price.data

    POWER_ENDPOINT = "/power/wem.json"

    write_output(POWER_ENDPOINT, stat_set, is_local=is_local)


def wem_export_daily(limit: int = None, is_local: bool = False):
    processed_years = 0

    for year in range(datetime.now().year, YEAR_MIN - 1, -1):

        stat_set = energy_fueltech_daily(year=year, network=NetworkWEM)

        weather = weather_daily(
            station_code="009021", year=year, network_code="WEM",
        )
        stat_set.data += weather.data

        write_output(
            f"/wem/energy/daily/{year}.json", stat_set, is_local=is_local
        )

        processed_years += 1

        if limit and limit >= processed_years:
            return None


def wem_export_monthly(is_local: bool = False):
    """

    """

    stat_set = energy_fueltech_daily(network=NetworkWEM, interval_size="1M")

    stat_set.data += weather_daily(
        station_code="009021", network_code="WEM",
    ).data

    write_output(
        "/wem/energy/monthly/all.json", stat_set, is_local=is_local,
    )


if __name__ == "__main__":
    if settings.env in ["development", "staging"]:
        export_power(priority=PriorityType.live)
        export_energy(latest=True)
        # export_metadata()
        # wem_export_power(is_local=True)
        # wem_export_daily(limit=1, is_local=True)
        # wem_export_monthly(is_local=True)
    else:
        wem_export_power()
        wem_export_daily()
        wem_export_monthly()



"""Optimized energy_sum sql function

Revision ID: 10f1d9517df7
Revises: 458bf49bd781
Create Date: 2020-10-29 03:25:01.093194

"""
from alembic import op

from opennem.core.loader import load_data

# revision identifiers, used by Alembic.
revision = "10f1d9517df7"
down_revision = "458bf49bd781"
branch_labels = None
depends_on = None


def upgrade():
    ENERGY_FUNCTIONS = load_data("energy_sum_v2.sql", from_fixture=True)
    op.execute(ENERGY_FUNCTIONS)


def downgrade():
    op.execute(
        """
        drop aggregate if exists energy_sum(numeric, text) cascade;

        drop type if exists agg_state_energy cascade;

        drop function if exists energy_sum_pop(agg_state_energy);

        drop function if exists interval_size(text, numeric);
    """
    )



import logging
import os

from requests import RequestException

from opennem.utils.handlers import open
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class LinkExtract(object):
    """
        parses and extracts links in items

    """

    @check_spider_pipeline
    def process_item(self, item, spider):
        if "link" not in item:
            return item

        url = item["link"]
        fh = None
        content = None
        _, file_extension = os.path.splitext(url)

        try:
            fh = open(url)
        except RequestException:
            logger.error("Bad link: {}".format(url))
        except Exception as e:
            logger.error("Error: {}".format(e))

        if fh:
            content = fh.read()

            item["content"] = content
            item["extension"] = file_extension

        return item



import logging
from datetime import datetime, timedelta

import requests

from opennem.core.networks import network_from_network_code
from opennem.monitors.aemo_wem_live_intervals import (
    get_aemo_wem_live_facility_intervals_recent_date,
)
from opennem.notifications.slack import slack_message
from opennem.utils.dates import parse_date

logger = logging.getLogger(__name__)


def get_wem_interval_delay() -> bool:
    resp = requests.get("https://data.opennem.org.au/power/wem.json")

    if resp.status_code != 200:
        logger.error("Error retrieving wem power")
        return False

    resp_json = resp.json()

    if "data" not in resp_json:
        logger.error("Error retrieving wem power: malformed response")
        return False

    data = resp_json["data"]
    network_code = resp_json["code"]

    network = network_from_network_code(network_code)

    fueltech_data = data.pop()

    history_end_date = fueltech_data["history"]["last"]

    history_date = parse_date(history_end_date, dayfirst=False)
    now_date = datetime.now().astimezone(network.get_timezone())

    live_most_recent = get_aemo_wem_live_facility_intervals_recent_date()

    time_delta = now_date - history_date

    if time_delta > timedelta(hours=3):
        slack_message(
            "WARNING: WEM live interval delay currently: {}\n\nFeed time: {}\nCurrent time: {}\n".format(
                time_delta, history_date, now_date
            )
        )

    live_delta = now_date - live_most_recent

    if live_delta > timedelta(minutes=90):
        slack_message(
            "AEMO Live intervals for WEM curently delayed by {}\n\nAEMO feed most recent: {}".format(
                live_delta, live_most_recent
            )
        )


if __name__ == "__main__":
    delay = get_wem_interval_delay()
